{-# OPTIONS -XRank2Types #-}

{-| This module provides: 

       * types for describing the implementation of certain operations.

       * functions that execute an implementation (the @X@ functions).

       * functions that @lift@ an implementation for a monad @m@ to 
         an implementation for a monad @t m@.
-}

module Monatron.Operations (
  Norm,
  -- * Make/with.
  MakeWith, MakeWithSig(..), liftMakeWith,
  withX, makeX, 
  -- ** Operations derived from Make/With.
  -- | The following operations use implementations of @MakeWith@.
  getX, collectX, putX, localX,
  -- * Throw and Handle
  Exc, ExcSig(..), liftExc, throwX, handleX,
  -- * Call-with-current-continuation.
  CallCC, liftCallCC, callCCX, 
  -- * MPlus.
  MPlus, MPlusSig(..), liftMPlus, mZeroX, mPlusX, 
  module Monatron.MonadT
) where

import Monatron.MonadT

-- | Type of 'correction' functions.
type Norm m = forall a. m a -> m a

--------------------------------------------------------

-- | Signature of MakeWith operations.
data MakeWithSig z a = With (z -> a)
                     | Make a z

-- | Type of implementations of MakeWith
type MakeWith z m = (forall a. MakeWithSig z a -> m a, Norm m) 

-- |Say what you would do with the context @z@ and @withX@ will do it.
withX :: Monad m => MakeWith z m -> (z -> m a) -> m a
withX (k,n) = n . join . k . With      

-- |Perform the computation in the context @z@.
makeX :: Monad m => MakeWith z m -> m a -> z -> m a
makeX (k,n)  m z = n $ join $ k $ Make m z

-- | Lift and implementation MakeWith to an arbitary monad transformer.
liftMakeWith :: (Monad m, MonadT t, Monad (t m)) => MakeWith z m -> MakeWith z (t m)
liftMakeWith (k,n) = (lift . k, tmap n)

-- derived:
-- | Get the current context @z@.
getX :: Monad m => MakeWith z m -> m z
getX op = withX op return

-- | Put @z@ in the context.
putX :: Monad m => MakeWith z m -> z -> m ()
putX op = makeX op (return ())

-- | Collect the current context.
collectX :: Monad m => MakeWith z m -> m a -> m (a,z)
collectX op m = m >>= \a -> withX op (\z -> return (a,z)) 

-- | Perform the computation @m@ in a context modified by @f@.
localX :: Monad m => MakeWith z m -> (z -> z) -> m a -> m a
localX op f m = do z <- getX op
                   makeX op m (f z)

--------------------------------------------------------
-- | Signature of Exc operations.
data ExcSig z a = Throw z
                | Handle a (z -> a)

-- | Type of implementations of Exc
type Exc z m = (forall a. ExcSig z a -> m a, Norm m)

-- | Throw and exception.
throwX :: Monad m => Exc z m -> z -> m a
throwX (k,n)  = n . k . Throw

-- | If the computation @m@ throws an exception, apply the handler.
handleX :: Monad m => Exc z m -> m a -> (z -> m a) -> m a
handleX (k,n) m  =  n . join . k . Handle m 

-- | Lift and implementation Exc to an arbitary monad transformer.
liftExc :: (Monad m, MonadT t, Monad (t m)) => Exc z m -> Exc z (t m)
liftExc (k,n)  = (lift . k, tmap n)

--------------------------------------------------------
-- | Type of implementations of CallCC
type CallCC m = (forall a b. ((a -> m b) -> a) -> m a, Norm m)

-- | Capture the current continuation.
callCCX :: (Monad m) => CallCC m -> ((a -> m b) -> m a) -> m a
callCCX (k,n) f = n $ join $ k (\c -> f (c . return))

-- | Lift and implementation CallCC to an arbitary monad transformer.
liftCallCC :: (Monad m, MonadT t, Monad (t m)) => CallCC m -> CallCC (t m)
liftCallCC (k,n)  = (\f -> lift $ k (\c -> f (lift. c)), tmap n)

--------------------------------------------------------
-- | Signature of MPlus operations.
data MPlusSig a = MZero | MPlus a a

-- | Type of implementations of MPlus
type MPlus m = (forall a. MPlusSig a -> m a, Norm m)

-- | The zero element of the monad.
mZeroX             :: Monad m => MPlus m -> m a
mZeroX (k,n)  = n (k MZero)

-- | Combines the effect of two computations.
mPlusX      :: Monad m => MPlus m -> m a -> m a -> m a
mPlusX (k,n)  m =  n . join . k . MPlus m

-- | Lift and implementation MPlus to an arbitary monad transformer.
liftMPlus :: (Monad m, MonadT t, Monad (t m)) => MPlus m -> MPlus (t m)
liftMPlus (k,n)  = (lift . k, tmap n)
