
-- | A collection of monad transformers and implementations of operations.

module Monatron.Transformer (
  -- * State monad transformer
  StateT, stateT, runStateT, makeWithStateT,
  -- * Writer monad transformer  
  WriterT, writerT, runWriterT, makeWithWriterT,
  -- * Reader monad transformer
  ReaderT, readerT, runReaderT, makeWithReaderT,
  -- * Exception monad transformer
  ExcT, excT, runExcT, excExcT,
  -- ** IO-exception monad
  IOExc, ioexc, runIOExc, excIOExc,
  -- * Continuation monad transformer
  ContT, contT, runContT, callCCContT,
  -- * List monad transformer
  ListT, LSig, listT, runListT, mPlusListT, foldListT, collectListT,
  mPlusList,
  module Monatron.Operations,
  module Data.Monoid
) where

import Monatron.Operations
import Control.Monad.Fix
import qualified Control.Exception as IO (throwIO,catch,Exception)
-- for Writer
import Data.Monoid
-- for Error (and Reader?)
import Monatron.Codensity

-- State Monad Transformer

-- | Adds support to @m@ for state of type @s@.
newtype StateT s m a = S { unS :: s -> m (a,s) }

stateT ::  (s -> m (a, s)) -> StateT s m a
stateT = S

runStateT :: s -> StateT s m a -> m (a,s) 
runStateT s m = unS m s

instance MonadT (StateT s) where
    lift  m           = S $ \s -> m >>= \a -> return (a,s)
    tmixmap f _ (S m) = S (f . m)
    m `tbind` k       = S $ \s -> unS m s >>= \ ~(a, s') -> unS (k a) s'

instance (Monad m) => Monad (StateT s m) where
    return = treturn
    fail   = lift . fail
    (>>=)  = tbind

instance (Monad m) => Functor (StateT s m) where fmap = liftM

instance (MonadFix m) => MonadFix (StateT s m) where
  mfix f  = S $ \s -> mfix (runStateT s . f . fst)

-- | @StateT@ implementation of operations 'MakeWith'
makeWithStateT :: Monad m => MakeWith s (StateT s m)
makeWithStateT = (k, id)
      where k (With f)   = S $ \s  -> return (f s, s)
            k (Make a s) = S $ \_ -> return (a,s)

--------------------------------------------------------------
-- Writer Monad Transformer

{-| Adds support to @m@ for collecting values of type @w@. 
   The type @w@ should be a monoid.
-}
newtype WriterT w m a = W {unW :: StateT w m a } 

writerT :: (Monoid w, Monad m) => m (a,w) -> WriterT w m a
writerT m = W $ S $ \w' -> m >>= \(a,w) -> return (a, w' `mappend` w)

runWriterT :: (Monoid w) => WriterT w m a -> m (a,w)
runWriterT x = unS (unW x) mempty
                  
instance Monoid w => MonadT (WriterT w) where
    lift         = W . lift
    tmixmap f g  = W . tmixmap f g . unW 
    m `tbind` f  = W (unW m >>= unW . f)

instance (Monad m, Monoid w) => Monad (WriterT w m) where
    return = treturn
    fail   = lift . fail
    (>>=)  = tbind

instance (Monad m, Monoid w) => Functor (WriterT w m) where fmap = liftM

instance (MonadFix m, Monoid w) => MonadFix (WriterT w m) where
    mfix f = W $ mfix (unW. f) 

-- | @WriterT@ implementation of operations 'MakeWith'
makeWithWriterT :: (Monoid w, Monad m) => MakeWith w (WriterT w m)
makeWithWriterT = (k, id)
        where k (With c)   = W $ S $ \w -> return (c w, mempty)
              k (Make a w) = writerT $ return (a,w)

--------------------------------------------------------------
-- Reader Monad Transformer
-- | Adds support to @m@ for environments of type @e@.
newtype ReaderT e m a = R {unR :: StateT e m a }

readerT :: Monad m => (e -> m a) -> ReaderT e m a
readerT r = R $ S $ \e -> r e >>= \a -> return (a,e)

runReaderT :: Monad m => e -> ReaderT e m a -> m a
runReaderT e m = (unS (unR m)) e >>= return . fst

instance MonadT (ReaderT e) where
    lift        = R . lift
    m `tbind` f = R (unR m >>= unR . f)
    tmixmap f g = R . tmixmap f g . unR 

instance (Monad m) => Monad (ReaderT e m) where
    return = treturn
    fail   = lift . fail
    (>>=)  = tbind

instance (Monad m) => Functor (ReaderT e m) where fmap = liftM

instance (MonadFix m) => MonadFix (ReaderT w m) where
    mfix f = R $ mfix (unR. f) 

-- | @ReaderT@ implementation of operations 'MakeWith'
makeWithReaderT :: Monad m => MakeWith e (ReaderT e m)
makeWithReaderT = (R. fst makeWithStateT, readerT. flip runReaderT)

--------------------------------------------------------------
-- Exceptions Monad Transformer

--Internal exceptions monad transformer
-- | Adds support to @m@ for throwing exceptions of type @x@.
newtype Xi x m a = Xi {unXi :: m (Either x a)}
--
{-
instance MonadT (Xi x) where
    lift m           = Xi (liftM Right m)
    tmixmap f _      = Xi . f . unXi
    (Xi m) `tbind` f = Xi (do a <- m
                              case a of
                                Left x  -> return (Left x)
                                Right b -> unXi (f b))

instance (Monad m) => Monad (Xi x m) where
    return = treturn
    fail   = lift . fail
    (>>=)  = tbind
-}

-- I don't want to export the Xi MonadT instance.
xilift :: (Monad m) => m a -> Xi x m a
xilift m           = Xi (liftM Right m)

xitmap :: (forall a. m (Either x a) -> n (Either x a)) -> Xi x m b -> Xi x n b
xitmap f        = Xi . f . unXi

instance (Monad m) => Monad (Xi x m) where
    return      = xilift. return
    fail        = xilift . fail
    Xi m >>= f  = Xi (do a <- m
                         case a of
                           Left x  -> return (Left x)
                           Right b -> unXi (f b))

{-
instance (Monad m) => Functor (Xi x m) where fmap = liftM
-}
instance (MonadFix m) => MonadFix (Xi x m) where
  mfix f  = Xi $ mfix (unXi . f . fromRight)
    where fromRight (Right a) = a
          fromRight _         = error "ExceptionT: mfix looped."

--
-- internal operations
throwi :: Monad m => x -> Xi x m a
throwi x = Xi $ return (Left x)
--
handlei :: Monad m => Xi x m a -> (x -> Xi x m a) -> Xi x m a
handlei m h = Xi (unXi m >>= \exa ->
                        case exa of
                          Left x  -> unXi (h x)
                          Right a -> return (Right a))
--
------------------------

-- External exceptions monad transformer
-- | Adds support to @m@ for exceptions of type @x@.
newtype ExcT x m a = X {unX :: Codensity (Xi x m) a}

excT :: Monad m => m (Either x a) -> ExcT x m a
excT = X. lift . Xi

runExcT :: Monad m => ExcT x m a -> m (Either x a)
runExcT = unXi . runCodensity . unX

instance MonadT (ExcT x) where
    lift        = X . lift . xilift
    tmixmap f g = X. tmixmap (xitmap f) (xitmap g) . unX
    m `tbind` f = X (unX m >>= unX . f)

instance (Monad m) => Monad (ExcT x m) where
    return = treturn
    fail   = lift . fail
    (>>=)  = tbind

instance (Monad m) => Functor (ExcT x m) where fmap = liftM

instance MonadFix m => MonadFix (ExcT x m) where
    mfix f =  X $ mfix (unX. f)

-- | @ExcT@ implementation of operations 'Exc'
excExcT :: Monad m => Exc x (ExcT x m)
excExcT = (k, excT. runExcT)
    where k (Throw x)    = X $ lift $ throwi x
          k (Handle a h) = X $ codensity $ \c -> handlei (c a) (c . h)

--------------------------------------------------------------
-- Monad for the IO implementation of exceptions.
-- | Monad implementing liftable IO exceptions.
newtype IOExc a = IOX {unIOX :: Codensity IO a}

ioexc :: IO a -> IOExc  a
ioexc = IOX. lift

runIOExc :: IOExc a -> IO a
runIOExc = runCodensity . unIOX

instance Monad IOExc where
    return  = IOX . return
    m >>= f = IOX (unIOX m >>= unIOX . f)

instance Functor IOExc where fmap = liftM

-- IOExc implementation of operations
-- | @IOExc@ implementation of operations 'Exc'
excIOExc :: Exc IO.Exception IOExc
excIOExc = (k, ioexc . runIOExc)
    where k (Throw x)    = IOX $ lift $ IO.throwIO x
          k (Handle a h) = IOX $ codensity $ \c -> IO.catch (c a) (c . h)


--------------------------------------------------------------
-- Continuations Monad Transformer

-- | Adds support to @m@ for continuations with result @r@.
newtype ContT r m a = C {unC :: (a -> m r) -> m r}

runContT :: (a -> m r) -> ContT r m a -> m r
runContT = flip unC

contT ::  ((a -> m r) -> m r) -> ContT r m a
contT = C

instance MonadT (ContT r) where
    lift m = C (m >>=)
    tmixmap f g m = C $ \k -> f $ unC m (g . k)
    m `tbind` k   = C $ \c -> unC m (\a -> unC (k a) c)

instance (Monad m) => Monad (ContT r m) where
    return = treturn
    fail   = lift . fail
    (>>=)  = tbind

instance (Monad m) => Functor (ContT r m) where fmap = liftM

-- | @ContT@ implementation of operation 'CallCC'
callCCContT :: Monad m => CallCC (ContT r m)
callCCContT = (k,id)
          where k f = C $ \c -> c (f (\a -> C $ \_ -> c a))

--------------------------------------------------------------
-- List monad transformer

-- | Adds support to @m@ for multiple answers.
data LSig f a b = NilT b
                | ConsT a (f a)

newtype ListT m a = L {unL :: m (LSig (ListT m) a ())}

runListT :: ListT m a -> m (LSig (ListT m) a ())
runListT = unL

listT :: m (LSig (ListT m) a ()) -> ListT m a
listT = L

emptyL :: Monad m => ListT m a
emptyL = L $ return $ NilT ()

appendL :: Monad m=> ListT m a -> ListT m a -> ListT m a
appendL (L m1) (L m2) = L $ do
            l <- m1
            case l of
              NilT ()    -> m2
              ConsT a l1 -> return (ConsT a (appendL l1 (L m2)))

foldListT :: Monad m => (a -> m b -> m b) -> m b -> ListT m a -> m b
foldListT c n (L m) = do l <- m 
                         case l of 
                            NilT ()    -> n 
                            ConsT a l1 -> c a (foldListT c n l1)

collectListT :: Monad m => ListT m a -> m [a]
collectListT lt = foldListT (\a m -> m >>= return. (a:)) (return []) lt

instance MonadT ListT where
    lift m            = L $ liftM (`ConsT` emptyL) m
    tmixmap f g (L m) = L $ f $  do 
                         l <- m
                         case l of
                            NilT ()    -> return (NilT ())
                            ConsT a l1 -> return $ ConsT a (tmixmap f g l1)
    m `tbind` f       = L $ foldListT (\a l -> unL $ f a `appendL` L l)
                                      (return $ NilT ())
                                      m

instance (Monad m) => Monad (ListT m) where
    return = treturn
    fail   = lift . fail
    (>>=)  = tbind

instance (Monad m) => Functor (ListT m) where fmap = liftM

-- | @ListT@ implementation of operations 'MPlus'
mPlusListT :: (Monad m) => MPlus (ListT m)
mPlusListT = (k,id)
          where k MZero       = emptyL
                k (MPlus a b) = appendL (return a) (return b)
------------
-- MPlus operation semantics for [] 
-- (in fact, for any well-behxaved MonadPlus)
-- | @[]@ implementation of operations 'MPlus'
mPlusList :: MPlus []
mPlusList = (k,id)
          where k MZero       = mzero
                k (MPlus a b) = mplus (return a) (return b)